#include <stdio.h>
#include "transform3d.hpp"
#include <stdlib.h>
#include "time_measurement.hpp"
#include "math_utils.hpp"
#include "cluster_registration.hpp"
#include <time.h>
#include "hashtable.hpp"
#include "point3d.hpp"
#include "proximity_verifier.hpp"
#include "config.hpp"
#include "density_clusterer.hpp"

#include "sampler.hpp"
#include "logger.hpp"

/* This function basically just performs principal component analysis to get decorrelating matrix */
Matrix3x3 find_rotation(Point3D* set, int size) {
	Matrix3x3 rot = matrix3x3_create();
	Matrix3x3 m = matrix3x3_create();

	/* Simplified convariance matrix. This simplifaction is possible because 
	 * points are already centered(this fact allows omitting subtraction from expected value)
	 * and accurate eigen VALUES aren't needed(we can omit all normalization) */
	for(int i = 0; i < size; ++i) {
		m.data[0] += set[i].x * set[i].x;
		m.data[1] += set[i].x * set[i].y;
		m.data[2] += set[i].x * set[i].z;
		m.data[4] += set[i].y * set[i].y;
		m.data[5] += set[i].y * set[i].z;
		m.data[8] += set[i].z * set[i].z;
	}

	/* Convariance matrix is by definition symmetrical, so to avoid redundantly computing 
	 * values we can just fill values under diagonal via symmetry*/
	m.data[3] = m.data[1];
	m.data[6] = m.data[2];
	m.data[7] = m.data[5];


	Matrix3x3 eigens;
	/* Beware! Might return different order of eigenvectors than reference registrator!
	 * (However when tested on same data it works). */
	evd(&m, &eigens);

	/* rx is 1st row of eigenvalue matrix */
	Point3D rx(eigens.data[0], eigens.data[1], eigens.data[2]);
	/* ry is 2nd row */
	Point3D ry(eigens.data[3], eigens.data[4], eigens.data[5]);
	Point3D rz = Point3D::crossproduct(rx, ry);

	/* Finally, put eigenvectors as columns */
	matrix3x3_put_to_column(&rot, rx, 0);
	matrix3x3_put_to_column(&rot, ry, 1);
	matrix3x3_put_to_column(&rot, rz, 2);

	return rot;
}

Transform3D build_best_candidate(Transform3D decorrelation, Transform3D best_match) {

	/* Because points are first rotated then translated, to merge transformations
	 * correctly, 'to_center' vector needs to be firstly rotated */
	Point3D rotated_translation(decorrelation.translation[0], decorrelation.translation[1], decorrelation.translation[2]);
	decorrelation.rotate_point_ref(rotated_translation);

	/* Actual decorrelating matrix */
	Transform3D rott(decorrelation.rotation, rotated_translation.coords);

	/* Final transformation is product of first applying decorrelation and then applying best match */
	return best_match * rott;
}

Transform3D cluster_registration_execute(TriangleMesh* mesh1, TriangleMesh* mesh2) {

	PROFILER_START_TIMER(RADIUS_INIT);
	LOG("Initializing radiuses for mesh #2");
	trianglemesh_init_radiuses(mesh2);
	LOG_FORMATTED("Mesh2:\n radius: %f\n weighted_radius: %f\n centroid: [%f; %f; %f]\n", mesh2->radius, mesh2->weighted_radius, mesh2->centroid.x, mesh2->centroid.y, mesh2->centroid.z);
	LOG_TIME_SPENT_SINCE(RADIUS_INIT, "Radiuses initialized in: ");

	PROFILER_START_TIMER(MESH_PREPROCESSING);
	LOG("Decorrelating mesh2...");

	/* Result of negating centroid is vector which after applying to
	* all the points centers whole mesh to [0; 0; 0] */
	Point3D to_center = mesh2->centroid;
	to_center.negate();

	/* Center mesh so its pointsum is zero */
	for(int i = 0; i < mesh2->points_size; ++i) {
		mesh2->points[i] += to_center;
	}

	Matrix3x3 rotation = find_rotation(mesh2->points, mesh2->points_size);
	LOG("Matrix of decorrelating rotation: ");
	LOG_DO_IF_LOGGER_ENABLED(print_matrix3x3(rotation));

	Transform3D decorrelating_transformation(rotation.data, to_center.coords);

	/* Transform mesh2 so its points are uncorrelated */
	for(int i = 0; i < mesh2->points_size; ++i) {
		decorrelating_transformation.rotate_point_ref(mesh2->points[i]);
		decorrelating_transformation.rotate_point_ref(mesh2->normals[i]);
	}

	LOG_TIME_SPENT_SINCE(MESH_PREPROCESSING, "Mesh2 decorrelated in: ");

	LOG("Initializing metric data...");
	metric_data_init(mesh2, mesh2->area_sum, mesh2->weighted_radius, mesh2->radius, config_rotation_coeff);

	LOG("Initializing proximity verifier...");
	PROFILER_START_TIMER(PROX_VER_INIT);

	ProximityVerifier* verifier = create_proximity_verifier(mesh1, mesh2, mesh2->radius * config_delta_ratio);
	LOG_FORMATTED("Proximity verifier:\n grid dimensions - [X: %d; Y: %d; Z: %d]\n step: %f, inverse-step: %f\n", verifier->grid_x, verifier->grid_y, verifier->grid_z, verifier->step, verifier->inverse_step);

	LOG_TIME_SPENT_SINCE(PROX_VER_INIT, "Proximity verifier initialized in: ");

	PROFILER_START_TIMER(CANDIDATES);
	int candidates_size;
	Transform3D* candidates = curvature_oracle_find_candidates(mesh1, mesh2, verifier, &candidates_size);
	LOG_TIME_SPENT_SINCE(CANDIDATES, "Space of candidate transformations generated in: ");

	density_clusterer_init(1, mesh2->radius);

	Transform3D best_candidate = density_clusterer_get_best_candidate(candidates, candidates_size);

	proximity_verifier_free(verifier);
	//free(candidates);

	/* The final transformation consist of applying decorrelation transformation
	 * and the best found transformation. */
	return build_best_candidate(decorrelating_transformation, best_candidate);
}
