
#include <time.h>
#include <stdio.h>
#include <thread>
#include <mutex>
#include <vector>
#include <stack>

#include "curvature_oracle.hpp"
#include "math_utils.hpp"

#include "time_measurement.hpp"
#include "logger.hpp"

static std::mutex candidates_mutex;
static std::vector<Transform3D> candidates;
static std::mutex curvatures_mutex;
static std::vector<PointRecord> curvatures;

std::vector<int> curvature_oracle_get_neighbours_euclid(int vertex_index, const NeighboursCache* neighbours_cache, float dist, const TriangleMesh* mesh) {

	std::vector<int> result;
	std::stack<int> stack;

	char* visited = (char*) calloc(mesh->points_size, sizeof(char));
	visited[vertex_index] = 1;

	float dist_squared = dist * dist;

	stack.push(vertex_index);
	result.push_back(vertex_index);

	Point3D center = mesh->points[vertex_index];

	/* Iterate while there still are items in the stack */
	while(!stack.empty()) {
		int curr_vertex = stack.top();
		stack.pop();
		
		CacheEntry cache_entry = neighbours_cache_get(neighbours_cache, curr_vertex);

		int neighbours_size = cache_entry.size;
		int* neighbours = cache_entry.data;

		/* Iterate over neighbours for curr_vertex */
		for(int i = 0; i < neighbours_size; ++i) {
			if(visited[neighbours[i]]) {
				continue;
			}

			/* If neighbour satisfy the condition of being 'dist' units away. */
			if(Point3D::distance_squared(mesh->points[neighbours[i]], center) <= dist_squared) {
				/* Add it to the result */
				result.push_back(neighbours[i]);

				/* It might have other points 'dist' units away in its neighbourhood. 
				 * Therefore add them to stack to be processed */
				stack.push(neighbours[i]);

				visited[neighbours[i]] = 1;
			}
		}
	}

	free(visited);

	return result;
}

int is_singular3x3(float* m) {

	float D = m[0] * m[2] - m[1] * m[1];

	if (fabs(D) < 0.0000000001) {
		return 1;
	}

	float D2 = m[0] * m[3];
	float y = D2 / D;

	if (fabs(m[3] * y - m[4]) < 0.000001) {
		return 1;
	}

	return 0;
}

void evd2x2(float* e1, float* e2, float* v1, float* v2, float a, float b, float c) {
	float d = sqrt((a - c) * (a - c) + 4 * b * b);
	*e1 = (a + c + d) / 2;
	*e2 = (a + c - d) / 2;

	if (fabs(*e2) < fabs(*e1)) {
		float tmp = *e2;
		*e2 = *e1;
		*e1 = tmp;
	}

	*v1 = 1;
	*v2 = 1;
	if (fabs(b) < fabs(a - *e1)) {
		*v1 = -b / (a - *e1);
	}
	else {
		*v2 = (*e1 - a) / b;
	}
}

void solve_symetric_stripe3x3(float a, float b, float c, float d, float e, float r[], float result[]) {
	float det = a * c * e - d * d * a - e * b * b;
	float det1 = r[0] * c * e + b * d * r[2] - d * d * r[0] - e * r[1] * b;
	float det2 = a * r[1] * e - r[2] * d * a - e * b * r[0];
	float det3 = a * c * r[2] + r[0] * b * d - d * r[1] * a - r[2] * b * b;

	result[0] = det1 / det;
	result[1] = det2 / det;
	result[2] = det3 / det;
}

int curvature_oracle_curvature_equality_condition(PointRecord p) {
	float ratio = p.k1 / p.k2;
	if(ratio < config_curvature_ratio) {
		return 1;
	}

	if(ratio > (1 / config_curvature_ratio)) {
		return 1;
	}

	return 0;
}

/* Returns:
 * 		1 - If it's possible to calculate curvatures
 * 		0 - If something fails.
 */
int curvature_oracle_compute_curvature(int vertex_index, const NeighboursCache* neighbours_cache, const TriangleMesh* mesh, PointRecord* result) {

	std::vector<int> neighbours = curvature_oracle_get_neighbours_euclid(vertex_index, neighbours_cache, mesh->avgEL * config_neighbourhood_width, mesh);

	if(neighbours.size() <= 2) {
		return 0;
	}

	Point3D point = mesh->points[vertex_index];
	Point3D v3 = mesh->normals[vertex_index];
	//Point3D v1 = POINTS_SUB(mesh->points[neighbours[1]], point);
	Point3D v1 = mesh->points[neighbours[1]] - point;

	/* v1 -= v3 * (v1 . v3) */
	//POINTS_SUB_REF(&v1, POINT_SCALE(v3, POINTS_DOTPRODUCT(v1, v3)));
	v1 -= v3 * Point3D::dotproduct(v1, v3);

	//if(POINT_ABS(v1) < 0.000001) {
	if(v1.abs() < 0.000001) {
		return 0;
	}

	//POINT_SCALE_REF(&v1, 1.0 / POINT_ABS(v1));
	v1 *= 1.0 / v1.abs();

	//Point3D v2 = POINTS_CROSSPRODUCT(v1, v3);
	Point3D v2 = Point3D::crossproduct(v1, v3);

	/* Matrix of rotation, we only need forst two rows */
	float rot[6];
	rot[0] = v1.x;
	rot[3] = v2.x;

	rot[1] = v1.y;
	rot[4] = v2.y;

	rot[2] = v1.z;
	rot[5] = v2.z;

	/* Rotated translation */
	float rtr[] = {
		-(rot[0] * point.x) - (rot[1] * point.y) - (rot[2] * point.z),
		-(rot[3] * point.x) - (rot[4] * point.y) - (rot[5] * point.z)
	};

	float m[5] = {0};
	float r[3] = {0};

	for(int i = 0; i < neighbours.size(); ++i) {
		int point_index = neighbours[i];

		float w = mesh->weights[point_index];
		Point3D p = mesh->points[point_index];
		Point3D normal = mesh->normals[point_index];

		float lpx = rot[0] * p.x + rot[1] * p.y + rot[2] * p.z + rtr[0];
		float lpy = rot[3] * p.x + rot[4] * p.y + rot[5] * p.z + rtr[1];
		float lnx = rot[0] * normal.x + rot[1] * normal.y + rot[2] * normal.z;
		float lny = rot[3] * normal.x + rot[4] * normal.y + rot[5] * normal.z;

		float wlpx = w * lpx;
		float wlpy = w * lpy;

		float wlpxlpx = wlpx * lpx;
		float wlpxlpy = wlpx * lpy;
		float wlpylpy = wlpy * lpy;


		m[0] += wlpxlpx;
		m[1] += wlpxlpy;
		m[2] += wlpylpy;
		r[0] += wlpx * lnx;
		r[1] += wlpy * lnx;

		m[2] += wlpxlpx;
		m[3] += wlpxlpy;
		m[4] += wlpylpy;
		r[1] += wlpx * lny;
		r[2] += wlpy * lny;
	}

	if(is_singular3x3(m)) {
		return 0;
	}

	float s[3];
	solve_symetric_stripe3x3(m[0], m[1], m[2], m[3], m[4], r, s);

	float e1;
	float e2;
	float egv1, egv2;

	evd2x2(&e1, &e2, &egv1, &egv2, s[0], s[1], s[2]);

	v1 *= egv1;
	v2 *= egv2;
	v1 += v2;
	v1 *= 1 / v1.abs();

	PointRecord point_record = point_record_create(e1, e2, v1, v3, point);
	*result = point_record;

	return 1;
}

/* Worker function - computes curvatures in specified range */
void* curvature_oracle_calculate_curvature_worker(CurvatureOraclePRArgs args, int thread_id) {

	int start = args.start;
	int n = args.n;
	const NeighboursCache* cache = args.cache;
	const TriangleMesh* mesh = args.mesh;
	int* samples = args.samples;

	PROFILER_START_TIMER(CURVATURE_COMPUTING);

	if(start + n > args.points_size) {
		LOG_FORMATTED("Potentional overflow detected(%d > %d), cutting expected amount of searched vertexes by %d\n", start + n, args.points_size, ((start + n) - args.points_size));
		n -= (start + n) - args.points_size;
	}

	for(int i = start; i < start + n; ++i) {
		PointRecord point_record;
		int ret = curvature_oracle_compute_curvature(samples[i], cache, mesh, &point_record);

		/* Computation of curvature succeeded */
		if(ret) {
			curvatures_mutex.lock();

			curvatures.push_back(point_record);

			curvatures_mutex.unlock();
		}
	}

	LOG_TIME_SPENT_SINCE(CURVATURE_COMPUTING, "Thread #" + std::to_string(thread_id) + " finished in: ");

	return NULL;
}

/* Prepares data for workers and starts them */
void curvature_oracle_start_curvature_workers(const TriangleMesh* mesh, int* sampled_mesh, int sampled_mesh_size) {

	PROFILER_START_TIMER(CTABLE_BUILD);
	CornerTable* ctable1 = corner_table_create(mesh);
	LOG_TIME_SPENT_SINCE(CTABLE_BUILD, "Corner table built in: ");

	NeighboursCache* cache1 = neighbours_cache_create(mesh->points_size, ctable1);

	int number_of_processors = config_number_of_processors;
	int to_process = ceil((float) sampled_mesh_size / number_of_processors);

	/* TODO: Caution, this might not work on compilers not allowing for
	 * VLA(variable length arrays), in case there's issue like this in the future
	 * replace this with malloc(calloc)/vector
	 *
	 * std::thread* threads = (std::thread*) malloc(sizeof(std::thread) * num); */
	std::thread threads[number_of_processors];

	CurvatureOraclePRArgs worker_args;
	/* Starting the threads */
	for(int i = 0; i < number_of_processors; ++i) {
		worker_args.start = i * to_process;
		worker_args.points_size = sampled_mesh_size;
		worker_args.n = to_process;
		worker_args.samples = sampled_mesh;
		worker_args.cache = cache1;
		worker_args.mesh = mesh;

		LOG_FORMATTED("Starting thread #%d with start in %dth vertex and searching through %d vertices.\n", i, i* to_process, to_process);
		threads[i] = std::thread(curvature_oracle_calculate_curvature_worker, worker_args, i);
	}

	/* Main thread waits until all of the workers finish */
	for(int i = 0; i < number_of_processors; ++i) {
		threads[i].join();
	}

	corner_table_free(ctable1);
	neighbours_cache_free(&cache1);
}



void* curvature_oracle_find_matches_worker(CurvatureOracleFMArgs args, int thread_id) {

	int start = args.start;
	int n = args.n;
	K2Tree* tree = args.tree;
	const NeighboursCache* cache = args.cache;
	const TriangleMesh* mesh = args.mesh;
	ProximityVerifier* verifier = (ProximityVerifier*) args.verifier;

	PROFILER_START_TIMER(OVERALL);

	if(start + n > args.candidates_size) {
		LOG_FORMATTED("Potentional overflow detected(%d > %d), cutting expected amount of searched vertexes by %d\n", start + n, args.candidates_size, ((start + n) - args.candidates_size));
		n -= (start + n) - args.candidates_size;
	}

	/* If PROFILING is defined when compiling, accumulate time spent by each step
	 * in generating transfromations */
#ifdef PROFILING
	std::chrono::microseconds CURVATURE_ACCUM = std::chrono::microseconds::zero();
	std::chrono::microseconds NEAREST_ACCUM = std::chrono::microseconds::zero();
	std::chrono::microseconds TRANSFORMATION_ACCUM = std::chrono::microseconds::zero();
	std::chrono::microseconds VERIFYING_ACCUM = std::chrono::microseconds::zero();
#endif
	
	for(int i = start; i < start + n; ++i) {

		PointRecord point_record;

		PROFILER_START_TIMER(CURVATURE);

		int curvature_success = curvature_oracle_compute_curvature(i, cache, mesh, &point_record);

#ifdef PROFILING
		CURVATURE_ACCUM += std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::high_resolution_clock::now() - CURVATURE);
#endif

		/* Computation of curvature failed */
		if(!curvature_success || !curvature_oracle_curvature_equality_condition(point_record)) {
			continue;
		}

		PROFILER_START_TIMER(NEAREST);
		/* Find point from the other mesh with nearest curvatures to this one */
		PointRecord nearest = k2tree_find_nearest(tree, point_record);
#ifdef PROFILING
		NEAREST_ACCUM += std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::high_resolution_clock::now() - NEAREST);
#endif
		PROFILER_START_TIMER(TRANSFORMATION);
		/* Construct transformation mapping points to each other */
		Transform3D transformation = transform3d_create_map(point_record, nearest, 0);
		Transform3D transformation_swapped = transform3d_create_map(point_record, nearest, 1);

		Transform3D transforms_to_test[] = {transformation, transformation_swapped};
#ifdef PROFILING
		TRANSFORMATION_ACCUM += std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::high_resolution_clock::now() - TRANSFORMATION);
#endif
		for(int j = 0; j < 2; ++j) {

			PROFILER_START_TIMER(VERIFYING);
			int verified = proximity_verifier_verify(verifier, transforms_to_test[j]);

			/* If transformation is verified as valid, add it to valid candidates */
			if(verified) {
				candidates_mutex.lock();
				
				candidates.push_back(transforms_to_test[j]);

				candidates_mutex.unlock();
			}
#ifdef PROFILING
			VERIFYING_ACCUM += std::chrono::duration_cast<std::chrono::microseconds>(std::chrono::high_resolution_clock::now() - VERIFYING);
#endif
		}

		/* If enough candidates were generated - end the thread
		 * Caution, this might result in more candidates being return than specified
		 * in config. */
		if(candidates.size() >= config_candidates_length) {
#ifdef PROFILING
			/* This is done via ugly string concating because there's race between threads
			 * for '<<' operator. */
			std::cout << "Thread #" + std::to_string(thread_id) + " ended prematurely after " + std::to_string(i - start) + " tries" \
			" in " + std::to_string(std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::high_resolution_clock::now() - OVERALL).count()) + "ms because enough candidates were found\n" \
			+ "  curvatures calculated in: " + std::to_string(CURVATURE_ACCUM.count() / 1000.0) + "ms\n" \
			+ "  nearest points found in: " + std::to_string(NEAREST_ACCUM.count() / 1000.0) + "ms\n" \
			+ "  transformations generated in: " + std::to_string(TRANSFORMATION_ACCUM.count() / 1000.0) + "ms\n" \
			+ "  transformations verified in: " + std::to_string(VERIFYING_ACCUM.count() / 1000.0) + "ms\n";
#endif
			return NULL;
		}
	}

#ifdef PROFILING
			/* This is done via ugly string concating because there's race between threads
			 * for '<<' operator. */
			std::cout << "Thread #" + std::to_string(thread_id) + " ended after expected " + std::to_string(n) + " tries" \
			" in " + std::to_string(std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::high_resolution_clock::now() - OVERALL).count()) + "ms\n" \
			+ "  curvatures calculated in: " + std::to_string(CURVATURE_ACCUM.count() / 1000.0) + "ms\n" \
			+ "  nearest points found in: " + std::to_string(NEAREST_ACCUM.count() / 1000.0) + "ms\n" \
			+ "  transformations generated in: " + std::to_string(TRANSFORMATION_ACCUM.count() / 1000.0) + "ms\n" \
			+ "  transformations verified in: " + std::to_string(VERIFYING_ACCUM.count() / 1000.0) + "ms\n";
#endif

	return NULL;
}

CandidatesTest* curvature_oracle_start_find_matches_workers(const TriangleMesh* mesh, const ProximityVerifier* verifier, int* sampled_mesh, int sampled_mesh_size, PointRecord* curvatures, int curvatures_size) {

	LOG("Finding candidate transformations...");

	PROFILER_START_TIMER(TREE_BUILD);
	K2Tree* tree = k2tree_create(curvatures, curvatures_size);
	LOG_TIME_SPENT_SINCE(TREE_BUILD, "K2Tree built in: ");

	PROFILER_START_TIMER(CTABLE_BUILD);
	CornerTable* ctable = corner_table_create(mesh);
	LOG_TIME_SPENT_SINCE(CTABLE_BUILD, "Corner table built in: ");

	NeighboursCache* cache = neighbours_cache_create(mesh->points_size, ctable);

	int number_of_processors = config_number_of_processors;
	int to_process = ceilf((float) sampled_mesh_size / number_of_processors);

	/* TODO: Caution, this might not work on compilers not allowing for
	 * VLA(variable length arrays), in case there's issue like this in the future
	 * replace this with malloc(calloc)/vector
	 *
	 * std::thread* threads = (std::thread*) malloc(sizeof(std::thread) * num); */
	std::thread threads[number_of_processors];

	CurvatureOracleFMArgs worker_args;
	/* Starting the threads */
	for(int i = 0; i < number_of_processors; ++i) {
		worker_args.start = i * to_process;
		worker_args.tree = tree;
		worker_args.n = to_process;
		worker_args.cache = cache;
		worker_args.mesh = mesh;
		worker_args.verifier = verifier;
		worker_args.candidates_size = sampled_mesh_size;

		LOG_FORMATTED("Starting thread #%d with start in %dth vertex and searching through %d vertices.\n", i, i * to_process, to_process);
		threads[i] = std::thread(curvature_oracle_find_matches_worker, worker_args, i);
	}

	/* Main thread waits until all of the workers finish */
	for(int i = 0; i < number_of_processors; ++i) {
		threads[i].join();
	}

	corner_table_free(ctable);
	neighbours_cache_free(&cache);
	k2tree_free(tree);

	return NULL;
}

void curvature_oracle_sample_meshes(TriangleMesh* mesh1, TriangleMesh* mesh2, int** sampled_mesh1, int* sampled_mesh1_size, int** sampled_mesh2, int* sampled_mesh2_size) {
	sampler_uniform_sampling_grid(mesh1, config_candidates_length, (mesh2->radius * config_delta_ratio) / 3.0, sampled_mesh1, sampled_mesh1_size);

	*sampled_mesh2 = (int*) malloc(sizeof(int) * mesh2->points_size);
	for(int i = 0; i < mesh2->points_size; ++i) {
		(*sampled_mesh2)[i] = i;
	}
	*sampled_mesh2_size = mesh2->points_size;
}

Transform3D* curvature_oracle_find_candidates(TriangleMesh* mesh1, TriangleMesh* mesh2, ProximityVerifier* verifier, int* candidates_size) {
	int* sampled_mesh1;
	int sampled_mesh1_size;

	int* sampled_mesh2;
	int sampled_mesh2_size;

	PROFILER_START_TIMER(SAMPLING);
	LOG("Sampling meshes...");
	curvature_oracle_sample_meshes(mesh1, mesh2, &sampled_mesh1, &sampled_mesh1_size, &sampled_mesh2, &sampled_mesh2_size);
	LOG_TIME_SPENT_SINCE(SAMPLING, "Meshes sampled in:");

	LOG("Starting workers finding curvatures for mesh #1...");
	curvature_oracle_start_curvature_workers(mesh1, sampled_mesh1, sampled_mesh1_size);

	free(sampled_mesh1);
	LOG_FORMATTED("Number of found points with valid curvatures: %ld\n", curvatures.size());

	LOG("Starting workers finding candidate transformations...");
	curvature_oracle_start_find_matches_workers(mesh2, verifier, sampled_mesh2, sampled_mesh2_size, &(curvatures[0]), curvatures.size());

	LOG_FORMATTED("Number of found candidate transformations: %ld\n", candidates.size());

	free(sampled_mesh2);

	*candidates_size = candidates.size();
	return &(candidates[0]);
}
