
#include <stddef.h>
#include <thread>
#include "density_clusterer.hpp"
#include "logger.hpp"
#include <vector>

/* TODO: This is awkward. Should be passed somehow instead. (DensityClustererContext) */
static float spread;
static float radius;

void* density_clusterer_calculate_density_worker(DCArgs args, int thread_id) {

	int start = args.start;
	int n = args.n;
	DistanceTree* tree = args.tree;
	Transform3D* candidates = args.candidates;
	int candidates_size = args.candidates_size;
	float* densities = args.densities;
	float threshold = args.threshold;

	PROFILER_START_TIMER(DENSITY_COMPUTING);

	if(start + n >= args.candidates_size) {
		n -= (start + n) - args.candidates_size;
	}

	for(int i = start; i < n + start; ++i) {
		std::vector<int> neighbours = distance_tree_find_close(tree, candidates[i], threshold);

		float density = 0;
		for(int j = 0; j < neighbours.size(); ++j) {
			density += exp((-(spread * spread) * metric_compute_squared(candidates[i], candidates[neighbours[j]])) / (radius * radius));
		}

		densities[i] = density;
	}

	LOG_TIME_SPENT_SINCE(DENSITY_COMPUTING, "Thread #" + std::to_string(thread_id) + " finished in: ");

	return NULL;
}

float* density_clusterer_start_workers(Transform3D* candidates, int candidates_size) {

	int number_of_processors = config_number_of_processors;
	int to_process = ceil((float) candidates_size / number_of_processors);
	float* densities = (float*) malloc(sizeof(float) * candidates_size);
	float threshold = -radius * log(0.01) / spread;

	PROFILER_START_TIMER(DISTANCE_TREE);
	DistanceTree* tree = distance_tree_create(candidates, candidates_size);
	LOG_TIME_SPENT_SINCE(DISTANCE_TREE, "Distance tree built in: ");

	/* TODO: Caution, this might not work on compilers not allowing for
	 * VLA(variable length arrays), in case there's issue like this in the future
	 * replace this with malloc(calloc)/vector
	 *
	 * std::thread* threads = (std::thread*) malloc(sizeof(std::thread) * num); */
	std::thread threads[number_of_processors];

	DCArgs worker_args;
	/* Starting the threads */
	for(int i = 0; i < number_of_processors; ++i) {
		worker_args.start = i * to_process;
		worker_args.n = to_process;
		worker_args.tree = tree;
		worker_args.candidates = candidates;
		worker_args.candidates_size = candidates_size;
		worker_args.threshold = threshold;
		worker_args.densities = densities;

		LOG_FORMATTED("Starting thread #%d with start in %dth vertex and searching through %d vertices.\n", i, i * to_process, to_process);
		threads[i] = std::thread(density_clusterer_calculate_density_worker, worker_args, i);
	}

	/* Main thread waits until all of the workers finish */
	for(int i = 0; i < number_of_processors; ++i) {
		threads[i].join();
	}

	distance_tree_free(tree);

	return densities;
}

Transform3D density_clusterer_get_best_candidate(Transform3D* candidates, int candidates_size) {
	LOG("Starting workers finding density for each transformation...");
	float* densities = density_clusterer_start_workers(candidates, candidates_size);

	float max_density = densities[0];
	int max_index = 0;

	for(int i = 1; i < candidates_size; ++i) {
		if(densities[i] > max_density) {
			max_density = densities[i];
			max_index = i;
		}
	}

	LOG_FORMATTED("Highest density observed: %f\n", max_density);
	
	free(densities);
	return candidates[max_index];
}

void density_clusterer_init(float _spread, float _radius) {
	spread = _spread;
	radius = _radius;
}
