#include <algorithm>

#include "distance_tree.hpp"


DistanceTreeNode* distance_tree_node_create() {
	DistanceTreeNode* node = (DistanceTreeNode*) malloc(sizeof(DistanceTreeNode));

	node->type = NON_LEAF;
	node->candidate_index = 0;
	node->close = NULL;
	node->far = NULL;
	node->children = NULL;
	node->children_size = 0;
	node->threshold = 0;

	return node;
}

int distance_tree_compare_float(const void* a, const void* b) {
	float pra = *((float*) a);
	float prb = *((float*) b);

	if(pra == prb) {
		return 0;
	}

	return pra > prb ? 1 : -1;
}

float distance_tree_find_threshold(DistanceTree* tree, int candidate_index, std::vector<int>& children) {
	float* distances = (float*) malloc(sizeof(float) * children.size());
	for(int i = 0; i < children.size(); ++i) {
		distances[i] = metric_compute(tree->candidates[candidate_index], tree->candidates[children[i]]);
	}

	std::nth_element(distances, &distances[children.size() / 2], &distances[children.size() - 1]);
	float median = distances[children.size() / 2];

	free(distances);

	return median;
}


DistanceTreeNode* distance_tree_build(DistanceTree* tree, int candidate_index, std::vector<int>& children) {
	DistanceTreeNode* node = distance_tree_node_create();

	node->candidate_index = candidate_index;

	/* In case there are less than 5 children. Mark node as leaf, add children and return */
	if(children.size() < DISTANCE_TREE_LEAF_CHILDREN_COUNT) {
		node->type = LEAF;
		node->children = (int*) malloc(sizeof(int) * children.size());
		memcpy(node->children, &children[0], sizeof(int) * children.size());
		node->children_size = children.size();

		return node;
	}
	
	float threshold = distance_tree_find_threshold(tree, candidate_index, children);
	node->threshold = threshold;

	std::vector<int> close;
	std::vector<int> far;

	/* TODO: Potentional optimalization - those values are already computed in distance_tree_find_threshold. */
	for(int i = 0; i < children.size(); ++i) {
		if(metric_compute(tree->candidates[candidate_index], tree->candidates[children[i]]) >= threshold) {
			far.push_back(children[i]);
		} else {
			close.push_back(children[i]);
		}
	}

	if(close.size() > 0) {
		/* Remove last candidate from array, and make node out of it */
		int close_index = close[close.size() - 1];
		close.pop_back();

		node->close = distance_tree_build(tree, close_index, close);
	}
	int far_index = far[far.size() - 1];
	far.pop_back();

	node->far = distance_tree_build(tree, far_index, far);

	return node;
}

void distance_tree_find_close_rec(DistanceTreeNode* node, DistanceTree* tree, Transform3D candidate, float threshold, std::vector<int>& vector) {
	float distance = metric_compute(candidate, tree->candidates[node->candidate_index]);
	if(distance <= threshold) {
		vector.push_back(node->candidate_index);
	}

	/* If node is a leaf add all children satisfying the condition of being max 'threshold' distance away and return */
	if(node->type == LEAF) {
		for(int i = 0; i < node->children_size; ++i) {
			if(metric_compute(candidate, tree->candidates[node->children[i]]) <= threshold) {
				vector.push_back(node->children[i]);
			}
		}
		return;
	}

	if((distance <= (node->threshold + threshold)) && node->close != NULL) {
		distance_tree_find_close_rec(node->close, tree, candidate, threshold, vector);
	}
	if((distance >= (node->threshold - threshold)) && node->far != NULL) {
		distance_tree_find_close_rec(node->far, tree, candidate, threshold, vector);
	}
}

std::vector<int> distance_tree_find_close(DistanceTree* tree, Transform3D candidate, float threshold) {
	std::vector<int> vector;

	distance_tree_find_close_rec(tree->root, tree, candidate, threshold, vector);

	return vector;
}

DistanceTree* distance_tree_create(Transform3D* candidates, int candidates_size) {
	DistanceTree* tree = (DistanceTree*) malloc(sizeof(DistanceTree));
	tree->candidates = candidates;
	tree->candidates_size = candidates_size;

	//int* candidates_indexes = (int*) malloc(sizeof(int) * (candidates_size - 1));
	std::vector<int> candidates_indexes;
	candidates_indexes.reserve(candidates_size);
	/* Iterating just to "size - 1" because 0 is reserved for root */
	for(int i = 0; i < candidates_size - 1; ++i) {
		candidates_indexes.push_back(i + 1);
	}

	tree->root = distance_tree_build(tree, 0, candidates_indexes);

	return tree;
}

void distance_tree_free_rec(DistanceTreeNode* node) {
	if(node->close != NULL) {
		distance_tree_free_rec(node->close);
	}

	if(node->far != NULL) {
		distance_tree_free_rec(node->far);
	}

	if(node->type == LEAF) {
		free(node->children);
	}

	free(node);
}

void distance_tree_free(DistanceTree* tree) {
	distance_tree_free_rec(tree->root);
	free(tree);
}
