#include <stdlib.h>
#include <stdio.h>
#include <stddef.h>
#include "trianglemesh.hpp"
#include "math_utils.hpp"

TriangleMesh* create_trianglemesh(Point3D* points, int points_size, Triangle* triangles, int triangles_size) {
	TriangleMesh* mesh = (TriangleMesh*) malloc(sizeof(TriangleMesh));

	if(mesh == NULL) {
		return NULL;
	}

	mesh->points = points;
	mesh->points_size = points_size;
	mesh->triangles = triangles;
	mesh->triangles_size = triangles_size;

	mesh->normals = NULL;
	mesh->radius = 0;
	mesh->max_area = 0;
	mesh->avgEL = 0;

	return mesh;
}

void trianglemesh_init_radiuses(TriangleMesh* mesh) {

	/* Computing radius */
	Point3D centroid(0, 0, 0);
	float sum = 0;

	for(int i = 0; i < mesh->points_size; ++i) {
		centroid += mesh->points[i];
	}

	centroid *= 1.0f / mesh->points_size;

	mesh->centroid = centroid;

	for(int i = 0; i < mesh->points_size; ++i) {
		sum += Point3D::distance(centroid, mesh->points[i]);
	}

	/* Radius is average distance from centroid */
	mesh->radius = sum / mesh->points_size;

	/* Computing weighted radius */
	Point3D weighted_centroid;
	Point3D average;
	sum = 0;
	for(int i = 0; i < mesh->triangles_size; ++i) {
		Point3D p1 = mesh->points[mesh->triangles[i].v1];
		Point3D p2 = mesh->points[mesh->triangles[i].v2];
		Point3D p3 = mesh->points[mesh->triangles[i].v3];

		average.x = (p1.x + p2.x + p3.x) / 3.0;
		average.y = (p1.y + p2.y + p3.y) / 3.0;
		average.z = (p1.z + p2.z + p3.z) / 3.0;

		float area = (Point3D::crossproduct(p2 - p1, p3 - p1)).abs() / 2;
		average *= area;
		weighted_centroid += average;
	}

	weighted_centroid *= 1.0f / mesh->area_sum;

	for(int i = 0; i < mesh->triangles_size; ++i) {
		Point3D p1 = mesh->points[mesh->triangles[i].v1];
		Point3D p2 = mesh->points[mesh->triangles[i].v2];
		Point3D p3 = mesh->points[mesh->triangles[i].v3];

		average.x = (p1.x + p2.x + p3.x) / 3.0;
		average.y = (p1.y + p2.y + p3.y) / 3.0;
		average.z = (p1.z + p2.z + p3.z) / 3.0;

		float area = (Point3D::crossproduct(p2 - p1, p3 - p1)).abs() / 2;
		sum += Point3D::distance(weighted_centroid, average) * area;

	}

	mesh->weighted_radius = sum / mesh->area_sum;
}

int init_trianglemesh(TriangleMesh* mesh) {
	Point3D* normals = (Point3D*) calloc(mesh->points_size, sizeof(Point3D));
	float* weights = (float*) calloc(mesh->points_size, sizeof(float));

	float avgEL = 0, max_area = 0, areasum = 0;

	if(normals == NULL || weights == NULL) {
		return EXIT_FAILURE;
	}

	/* TODO: Test performance by removing if statement in the loop and
	 * replacing it with storing the scalar instead
	 * TODO: TriangleMesh.weightedRadius() should be only computed for 
	 * one of the meshes, test performance by excluding it from here */

	for(int i = 0; i < mesh->triangles_size; ++i) {
		Triangle t = mesh->triangles[i];

		Point3D p1 = mesh->points[t.v1];
		Point3D p2 = mesh->points[t.v2];
		Point3D p3 = mesh->points[t.v3];

		/* Vectors p1->p2, p1->p3, p2->p3
		 * so basically vectors representing edges */
		Point3D v1 = p2 - p1;
		Point3D v2 = p3 - p1;
		Point3D v3 = p2 - p3;

		Point3D cross = Point3D::crossproduct(v1, v2);

		float cross_abs = cross.abs();
		float area = 0.5 * cross_abs;

		areasum += area;

		/* Computing weights which corresponds to area around vertex */
		float triangle_area = area / 3;
		weights[t.v1] += triangle_area;
		weights[t.v2] += triangle_area;
		weights[t.v3] += triangle_area;

		/* Sum of lenghts of all edges */
		avgEL += v1.abs();
		avgEL += v2.abs();
		avgEL += v3.abs();

		/* Computing vertex normals
		 * Accumulating normalized cross product of all triangles adjacent to vertex */
		if(cross_abs != 0) {
		/*if(cross_abs > EPSILON) {*/
			cross *= 1 / cross_abs;
			                                         
			normals[t.v1] += cross;
			normals[t.v2] += cross;
			normals[t.v3] += cross;
		}
	}

	mesh->area_sum = areasum;

	areasum /= mesh->points_size;
	avgEL /= mesh->triangles_size * 3;

	for(int i = 0; i < mesh->points_size; ++i) {

		weights[i] /= areasum;

		if(weights[i] > max_area) {
			max_area = weights[i];
		}

		float normal_abs = normals[i].abs();

		if(normal_abs > EPSILON) {
			normals[i] *= 1 / normal_abs;
		} else {
			normals[i].x = 1;
			normals[i].y = 0;
			normals[i].z = 0;
		}
	}

	mesh->normals = normals;
	mesh->weights = weights;
	mesh->avgEL = avgEL;
	mesh->max_area = max_area;

	return EXIT_SUCCESS;
}


void trianglemesh_free(TriangleMesh* mesh) {
	free(mesh->points);
	free(mesh->weights);
	free(mesh->triangles);
	free(mesh->normals);
	free(mesh);
}
